@@ -2,8 +2,6 @@
 
 import locale
 import os
-import re
-from signal import SIGINT, SIGTERM
 import struct
 from subprocess import Popen, PIPE
 import sys
@@ -25,13 +23,18 @@ try:
 except ImportError:
     termios = None
 
-from .exceptions import Failure, ThreadException
+from .exceptions import (
+    UnexpectedExit, Failure, ThreadException, WatcherError,
+)
 from .platform import (
-    WINDOWS, pty_size, character_buffered, ready_for_reading, read_byte,
+    WINDOWS, pty_size, character_buffered, ready_for_reading, bytes_to_read,
 )
 from .util import has_fileno, isatty, ExceptionHandlingThread
 
-from .vendor import six
+try:
+    from .vendor import six
+except ImportError:
+    import six
 
 
 class Runner(object):
@@ -83,9 +86,9 @@ class Runner(object):
         self.input_sleep = self.__class__.input_sleep
         #: Whether pty fallback warning has been emitted.
         self.warned_about_pty_fallback = False
-        #: The trigger/response mapping for use by `respond`. Is filled in at
-        #: runtime by `run`.
-        self.responses = None
+        #: A list of `.StreamWatcher` instances for use by `respond`. Is filled
+        #: in at runtime by `run`.
+        self.watchers = []
 
     def run(self, command, **kwargs):
         """
@@ -103,9 +106,19 @@ class Runner(object):
         :param str shell: Which shell binary to use. Default: ``/bin/bash``.
 
         :param bool warn:
-            Whether to warn and continue, instead of raising `.Failure`, when
-            the executed command exits with a nonzero status. Default:
-            ``False``.
+            Whether to warn and continue, instead of raising
+            `.UnexpectedExit`, when the executed command exits with a
+            nonzero status. Default: ``False``.
+
+            .. note::
+                This setting has no effect on exceptions, which will still be
+                raised, typically bundled in `.ThreadException` objects if they
+                were raised by the IO worker threads.
+
+                Similarly, `.WatcherError` exceptions raised by
+                `.StreamWatcher` instances will also ignore this setting, and
+                will usually be bundled inside `.Failure` objects (in order to
+                preserve the execution context).
 
         :param hide:
             Allows the caller to disable ``run``'s default behavior of copying
@@ -176,7 +189,7 @@ class Runner(object):
             of `default_encoding`).
 
         :param out_stream:
-            A file-like stream object to which the subprocess' standard error
+            A file-like stream object to which the subprocess' standard output
             should be written. If ``None`` (the default), ``sys.stdout`` will
             be used.
 
@@ -188,16 +201,15 @@ class Runner(object):
             A file-like stream object to used as the subprocess' standard
             input. If ``None`` (the default), ``sys.stdin`` will be used.
 
-        :param dict responses:
-            A `dict` whose keys are regular expressions to be searched for in
-            the program's ``stdout`` or ``stderr``, and whose values may be any
-            value one desires to write into a stdin text/binary stream
+        :param list watchers:
+            A list of `.StreamWatcher` instances which will be used to scan the
+            program's ``stdout`` or ``stderr`` and may write into its ``stdin``
             (typically ``str`` or ``bytes`` objects depending on Python
-            version) in response.
+            version) in response to patterns or other heuristics.
 
-            See :doc:`/concepts/responses` for details on this functionality.
+            See :doc:`/concepts/watchers` for details on this functionality.
 
-            Default: ``{}``.
+            Default: ``[]``.
 
         :param bool echo_stdin:
             Whether to write data from ``in_stream`` back to ``out_stream``.
@@ -231,22 +243,24 @@ class Runner(object):
         :returns:
             `Result`, or a subclass thereof.
 
-        :raises: `.Failure`, if the command exited nonzero & ``warn=False``.
+        :raises:
+            `.UnexpectedExit`, if the command exited nonzero and
+            ``warn`` was ``False``.
 
         :raises:
-            `.ThreadException` (if the background I/O threads encounter
-            exceptions).
+            `.Failure`, if the command didn't even exit cleanly, e.g. if a
+            `.StreamWatcher` raised `.WatcherError`.
 
         :raises:
-            ``KeyboardInterrupt``, if the user generates one during command
-            execution by pressing Ctrl-C.
-
-            .. note::
-                In normal usage, Invoke's top-level CLI tooling will catch
-                these & exit with return code ``130`` (typical POSIX behavior)
-                instead of printing a traceback and exiting ``1`` (which is
-                what Python normally does).
+            `.ThreadException` (if the background I/O threads encountered
+            exceptions other than `.WatcherError`).
         """
+        try:
+            return self._run_body(command, **kwargs)
+        finally:
+            self.stop()
+
+    def _run_body(self, command, **kwargs):
         # Normalize kwargs w/ config
         opts, out_stream, err_stream, in_stream = self._run_opts(kwargs)
         shell = opts['shell']
@@ -264,7 +278,7 @@ class Runner(object):
         thread_args = {
             self.handle_stdout: {
                 'buffer_': stdout,
-                'hide': 'out' in opts['hide'],
+                'hide': 'stdout' in opts['hide'],
                 'output': out_stream,
             },
             # TODO: make this & related functionality optional, for users who
@@ -281,62 +295,64 @@ class Runner(object):
         if not self.using_pty:
             thread_args[self.handle_stderr] = {
                 'buffer_': stderr,
-                'hide': 'err' in opts['hide'],
+                'hide': 'stderr' in opts['hide'],
                 'output': err_stream,
             }
         # Kick off IO threads
-        self.threads, exceptions = [], []
+        self.threads = {}
+        exceptions = []
         for target, kwargs in six.iteritems(thread_args):
             t = ExceptionHandlingThread(target=target, kwargs=kwargs)
-            self.threads.append(t)
+            self.threads[target] = t
             t.start()
         # Wait for completion, then tie things off & obtain result
         # And make sure we perform that tying off even if things asplode.
         exception = None
-        try:
-            self.wait()
-        except BaseException as e: # Make sure we nab ^C etc
-            exception = e
-            # TODO: consider consuming the KeyboardInterrupt instead of storing
-            # it for later raise; this would allow for subprocesses which don't
-            # actually exit on Ctrl-C (e.g. vim). NOTE: but this would make it
-            # harder to correctly detect it and exit 130 once everything wraps
-            # up...
-            # TODO: generally, but especially if we do ignore
-            # KeyboardInterrupt, honor other signals sent to our own process
-            # and transmit them to the subprocess before handling 'normally'.
-            # TODO: we should probably re-raise anything that's not
-            # KeyboardInterrupt? This is quite possibly swallowing things.
-            # NOTE: we handle this now instead of at actual-exception-handling
-            # time because otherwise the stdout/err reader threads may block
-            # until the subprocess exits.
-            if isinstance(exception, KeyboardInterrupt):
-                self.send_interrupt(exception)
+        while True:
+            try:
+                self.wait()
+                break # done waiting!
+            # NOTE: we handle all this now instead of at
+            # actual-exception-handling time because otherwise the stdout/err
+            # reader threads may block until the subprocess exits.
+            # TODO: honor other signals sent to our own process and transmit
+            # them to the subprocess before handling 'normally'.
+            except KeyboardInterrupt as e:
+                self.send_interrupt(e)
+                # NOTE: no break; we want to return to self.wait()
+            except BaseException as e: # Want to handle SystemExit etc still
+                # Store exception for post-shutdown reraise
+                exception = e
+                # Break out of return-to-wait() loop - we want to shut down
+                break
+        # Inform stdin-mirroring worker to stop its eternal looping
         self.program_finished.set()
-        for t in self.threads:
-            # NOTE: using a join timeout for corner case from #351 (one pipe
-            # excepts, fills up, prevents subproc from exiting, and other pipe
-            # then has a blocking read() call, causing its thread to block on
-            # join). In normal, non-#351 situations this should function
-            # similarly to a non-timeout'd join.
-            # NOTE: but we avoid a timeout for the stdin handler as it has its
-            # own termination conditions & isn't subject to this corner case.
-            timeout = None
-            if t.kwargs['target'] != self.handle_stdin:
-                # TODO: make the timeout configurable
-                timeout = 1
-            t.join(timeout)
-            e = t.exception()
+        # Join threads, setting a timeout if necessary
+        for target, thread in six.iteritems(self.threads):
+            thread.join(self._thread_timeout(target))
+            e = thread.exception()
             if e is not None:
                 exceptions.append(e)
         # If we got a main-thread exception while wait()ing, raise it now that
         # we've closed our worker threads.
         if exception is not None:
             raise exception
+        # Strip out WatcherError from any thread exceptions; they are bundled
+        # into Failure handling at the end.
+        watcher_errors = []
+        thread_exceptions = []
+        for exception in exceptions:
+            real = exception.value
+            if isinstance(real, WatcherError):
+                watcher_errors.append(real)
+            else:
+                thread_exceptions.append(exception)
         # If any exceptions appeared inside the threads, raise them now as an
         # aggregate exception object.
-        if exceptions:
-            raise ThreadException(exceptions)
+        if thread_exceptions:
+            raise ThreadException(thread_exceptions)
+        # At this point, we had enough success that we want to be returning or
+        # raising detailed info about our execution; so we generate a Result.
         stdout = ''.join(stdout)
         stderr = ''.join(stderr)
         if WINDOWS:
@@ -347,9 +363,13 @@ class Runner(object):
             # it is likely to matter for these days.
             stdout = stdout.replace("\r\n", "\n").replace("\r", "\n")
             stderr = stderr.replace("\r\n", "\n").replace("\r", "\n")
-        # Get return/exit code
-        exited = self.returncode()
-        # Return, or raise as failure, our final result
+        # Get return/exit code, unless there were WatcherErrors to handle.
+        # NOTE: In that case, returncode() may block waiting on the process
+        # (which may be waiting for user input). Since most WatcherError
+        # situations lack a useful exit code anyways, skipping this doesn't
+        # really hurt any.
+        exited = None if watcher_errors else self.returncode()
+        # Obtain actual result
         result = self.generate_result(
             command=command,
             shell=shell,
@@ -358,9 +378,17 @@ class Runner(object):
             stderr=stderr,
             exited=exited,
             pty=self.using_pty,
+            hide=opts['hide'],
         )
+        # Any presence of WatcherError from the threads indicates a watcher was
+        # upset and aborted execution; make a generic Failure out of it and
+        # raise that.
+        if watcher_errors:
+            # TODO: ambiguity exists if we somehow get WatcherError in *both*
+            # threads...as unlikely as that would normally be.
+            raise Failure(result, reason=watcher_errors[0])
         if not (result or opts['warn']):
-            raise Failure(result)
+            raise UnexpectedExit(result)
         return result
 
     def _run_opts(self, kwargs):
@@ -375,7 +403,11 @@ class Runner(object):
         for key, value in six.iteritems(self.context.config.run):
             runtime = kwargs.pop(key, None)
             opts[key] = value if runtime is None else runtime
-        # TODO: handle invalid kwarg keys (anything left in kwargs)
+        # Handle invalid kwarg keys (anything left in kwargs).
+        # Act like a normal function would, i.e. TypeError
+        if kwargs:
+            err = "run() got an unexpected keyword argument '{0}'"
+            raise TypeError(err.format(list(kwargs.keys())[0]))
         # If hide was True, turn off echoing
         if opts['hide'] is True:
             opts['echo'] = False
@@ -394,11 +426,26 @@ class Runner(object):
             in_stream = sys.stdin
         # Determine pty or no
         self.using_pty = self.should_use_pty(opts['pty'], opts['fallback'])
-        # Responses
-        # TODO: precompile the keys into regex objects
-        self.responses = opts.get('responses', {})
+        if opts['watchers']:
+            self.watchers = opts['watchers']
         return opts, out_stream, err_stream, in_stream
 
+    def _thread_timeout(self, target):
+        # Add a timeout to out/err thread joins when it looks like they're not
+        # dead but their counterpart is dead; this indicates issue #351 (fixed
+        # by #432) where the subproc may hang because its stdout (or stderr) is
+        # no longer being consumed by the dead thread (and a pipe is filling
+        # up.) In that case, the non-dead thread is likely to block forever on
+        # a `recv` unless we add this timeout.
+        if target == self.handle_stdin:
+            return None
+        opposite = self.handle_stderr
+        if target == self.handle_stderr:
+            opposite = self.handle_stdout
+        if opposite in self.threads and self.threads[opposite].is_dead:
+            return 1
+        return None
+
     def generate_result(self, **kwargs):
         """
         Create & return a suitable `Result` instance from the given ``kwargs``.
@@ -471,7 +518,7 @@ class Runner(object):
         stream.write(string)
         stream.flush()
 
-    def _handle_output(self, buffer_, hide, output, reader, indices):
+    def _handle_output(self, buffer_, hide, output, reader):
         # TODO: store un-decoded/raw bytes somewhere as well...
         for data in self.read_proc_output(reader):
             # Echo to local stdout if necessary
@@ -486,8 +533,8 @@ class Runner(object):
             # NOTE: this is threadsafe insofar as no reading occurs until after
             # the thread is join()'d.
             buffer_.append(data)
-            # Run our specific buffer & indices through the autoresponder
-            self.respond(buffer_, indices)
+            # Run our specific buffer through the autoresponder framework
+            self.respond(buffer_)
 
     def handle_stdout(self, buffer_, hide, output):
         """
@@ -509,7 +556,6 @@ class Runner(object):
             hide,
             output,
             reader=self.read_proc_stdout,
-            indices=threading.local(),
         )
 
     def handle_stderr(self, buffer_, hide, output):
@@ -524,12 +570,11 @@ class Runner(object):
             hide,
             output,
             reader=self.read_proc_stderr,
-            indices=threading.local(),
         )
 
     def read_our_stdin(self, input_):
         """
-        Read & decode one byte from a local stdin stream.
+        Read & decode bytes from a local stdin stream.
 
         :param input_:
             Actual stream object to read from. Maps to ``in_stream`` in `run`,
@@ -537,23 +582,23 @@ class Runner(object):
             object.
 
         :returns:
-            A Unicode string, the result of decoding the read byte (this might
+            A Unicode string, the result of decoding the read bytes (this might
             be the empty string if the pipe has closed/reached EOF); or
             ``None`` if stdin wasn't ready for reading yet.
         """
         # TODO: consider moving the character_buffered contextmanager call in
         # here? Downside is it would be flipping those switches for every byte
         # read instead of once per session, which could be costly (?).
-        byte = None
+        bytes_ = None
         if ready_for_reading(input_):
-            byte = read_byte(input_)
+            bytes_ = input_.read(bytes_to_read(input_))
             # Decode if it appears to be binary-type. (From real terminal
             # streams, usually yes; from file-like objects, often no.)
-            if byte and isinstance(byte, six.binary_type):
+            if bytes_ and isinstance(bytes_, six.binary_type):
                 # TODO: will decoding 1 byte at a time break multibyte
                 # character encodings? How to square interactivity with that?
-                byte = self.decode(byte)
-        return byte
+                bytes_ = self.decode(bytes_)
+        return bytes_
 
     def handle_stdin(self, input_, output, echo):
         """
@@ -577,25 +622,29 @@ class Runner(object):
 
         :returns: ``None``.
         """
+        # TODO: reinstate lock/whatever thread logic from fab v1 which prevents
+        # reading from stdin while other parts of the code are prompting for
+        # runtime passwords? (search for 'input_enabled')
+        # TODO: fabric#1339 is strongly related to this, if it's not literally
+        # exposing some regression in Fabric 1.x itself.
         with character_buffered(input_):
             while True:
-                # Read 1 byte at a time for interactivity's sake.
-                char = self.read_our_stdin(input_)
-                if char:
+                data = self.read_our_stdin(input_)
+                if data:
                     # Mirror what we just read to process' stdin.
                     # We perform an encode so Python 3 gets bytes (streams +
                     # str's in Python 3 == no bueno) but skip the decode step,
                     # since there's presumably no need (nobody's interacting
                     # with this data programmatically).
-                    self.write_proc_stdin(char)
+                    self.write_proc_stdin(data)
                     # Also echo it back to local stdout (or whatever
                     # out_stream is set to) when necessary.
                     if echo is None:
                         echo = self.should_echo_stdin(input_, output)
                     if echo:
-                        self.write_our_output(stream=output, string=char)
+                        self.write_our_output(stream=output, string=data)
                 # Empty string/char/byte != None. Can't just use 'else' here.
-                elif char is not None:
+                elif data is not None:
                     # When reading from file-like objects that aren't "real"
                     # terminal streams, an empty byte signals EOF.
                     break
@@ -603,31 +652,11 @@ class Runner(object):
                 # running, *and* we don't seem to be reading anything out of
                 # stdin. (NOTE: If we only test the former, we may encounter
                 # race conditions re: unread stdin.)
-                if self.program_finished.is_set() and not char:
+                if self.program_finished.is_set() and not data:
                     break
                 # Take a nap so we're not chewing CPU.
                 time.sleep(self.input_sleep)
 
-        # while not self.program_finished.is_set():
-        #    # TODO: reinstate lock/whatever thread logic from fab v1 which
-        #    # prevents reading from stdin while other parts of the code are
-        #    # prompting for runtime passwords? (search for 'input_enabled')
-        #    if have_char and chan.input_enabled:
-        #        # Send all local stdin to remote end's stdin
-        #        #byte = msvcrt.getch() if WINDOWS else sys.stdin.read(1)
-        #        yield self.encode(sys.stdin.read(1))
-        #        # Optionally echo locally, if needed.
-        #        # TODO: how to truly do this? access the out_stream which
-        #        # isn't currently visible to us? if we just skip this part,
-        #        # interactive users may not have their input echoed...ISTR we
-        #        # used to assume remote would send it back down stdout/err...
-        #        # clearly not?
-        #        #if not using_pty and env.echo_stdin:
-        #            # Not using fastprint() here -- it prints as 'user'
-        #            # output level, don't want it to be accidentally hidden
-        #        #    sys.stdout.write(byte)
-        #        #    sys.stdout.flush()
-
     def should_echo_stdin(self, input_, output):
         """
         Determine whether data read from ``input_`` should echo to ``output``.
@@ -640,54 +669,29 @@ class Runner(object):
         """
         return (not self.using_pty) and isatty(input_)
 
-    def respond(self, buffer_, indices):
+    def respond(self, buffer_):
         """
         Write to the program's stdin in response to patterns in ``buffer_``.
 
-        The patterns and responses are driven by the key/value pairs in the
-        ``responses`` kwarg of `run` - see its documentation for format
-        details, and :doc:`/concepts/responses` for a conceptual overview.
+        The patterns and responses are driven by the `.StreamWatcher` instances
+        from the ``watchers`` kwarg of `run` - see :doc:`/concepts/watchers`
+        for a conceptual overview.
 
         :param list buffer:
             The capture buffer for this thread's particular IO stream.
 
-        :param indices:
-            A `threading.local` object upon which is (or will be) stored the
-            last-seen index for each key in ``responses``. Allows the responder
-            functionality to be used by multiple threads (typically, one each
-            for stdout and stderr) without conflicting.
-
         :returns: ``None``.
         """
-        # Short-circuit if there are no responses to respond to. This saves us
-        # the effort of joining the buffer and so forth.
-        if not self.responses:
-            return
-        # Join buffer contents into a single string; without this, we can't be
-        # sure that the pattern we seek isn't split up across chunks in the
-        # buffer.
+        # Join buffer contents into a single string; without this,
+        # StreamWatcher subclasses can't do things like iteratively scan for
+        # pattern matches.
         # NOTE: using string.join should be "efficient enough" for now, re:
-        # speed and memory use. Should that turn up false, consider using
-        # StringIO or cStringIO (tho the latter doesn't do Unicode well?)
-        # which is apparently even more efficient.
+        # speed and memory use. Should that become false, consider using
+        # StringIO or cStringIO (tho the latter doesn't do Unicode well?) which
+        # is apparently even more efficient.
         stream = u''.join(buffer_)
-        # Initialize seek indices
-        if not hasattr(indices, 'seek'):
-            indices.seek = {}
-            for pattern in self.responses:
-                indices.seek[pattern] = 0
-        for pattern, response in six.iteritems(self.responses):
-            # Only look at stream contents we haven't seen yet, to avoid dupes.
-            new_ = stream[indices.seek[pattern]:]
-            # Search, across lines if necessary
-            matches = re.findall(pattern, new_, re.S)
-            # Update seek index if we've matched
-            if matches:
-                indices.seek[pattern] += len(new_)
-            # Iterate over findall() response in case >1 match occurred.
-            for _ in matches:
-                # TODO: automatically append system-appropriate newline if
-                # response doesn't end with it, w/ option to disable?
+        for watcher in self.watchers:
+            for response in watcher.submit(stream):
                 self.write_proc_stdin(response)
 
     def generate_env(self, env, replace_env):
@@ -729,7 +733,7 @@ class Runner(object):
             ``True`` if any threads appear to have terminated with an
             exception, ``False`` otherwise.
         """
-        return any(x.is_dead for x in self.threads)
+        return any(x.is_dead for x in self.threads.values())
 
     def wait(self):
         """
@@ -848,16 +852,35 @@ class Runner(object):
         """
         Submit an interrupt signal to the running subprocess.
 
+        In almost all implementations, the default behavior is what will be
+        desired: submit ``\x03`` to the subprocess' stdin pipe. However, we
+        leave this as a public method in case this default needs to be
+        augmented or replaced.
+
         :param interrupt:
             The locally-sourced ``KeyboardInterrupt`` causing the method call.
 
         :returns: ``None``.
         """
-        raise NotImplementedError
+        self.write_proc_stdin(u'\x03')
 
     def returncode(self):
         """
         Return the numeric return/exit code resulting from command execution.
+
+        :returns: `int`
+        """
+        raise NotImplementedError
+
+    def stop(self):
+        """
+        Perform final cleanup, if necessary.
+
+        This method is called within a ``finally`` clause inside the main `run`
+        method. Depending on the subclass, it may be a no-op, or it may do
+        things such as close network connections or open files.
+
+        :returns: ``None``
         """
         raise NotImplementedError
 
@@ -981,36 +1004,73 @@ class Local(Runner):
         else:
             return self.process.poll() is not None
 
-    def send_interrupt(self, interrupt):
-        # NOTE: No need to reraise the interrupt since we have full control
-        # over the local process and can kill it.
-        if self.using_pty:
-            os.kill(self.pid, SIGINT)
-        else:
-            # Use send_signal with platform-appropriate signal (Windows doesn't
-            # support SIGINT unfortunately, only SIGTERM).
-            # NOTE: could use subprocess.terminate() (which is cross-platform)
-            # but feels best to use SIGINT as much as we possibly can as it's
-            # most appropriate. terminate() always sends SIGTERM.
-            # NOTE: in interactive POSIX terminals, this is technically
-            # unnecessary as Ctrl-C submits the INT to the entire foreground
-            # process group (which will be both Invoke and its spawned
-            # subprocess). However, it doesn't seem to hurt, & ensures that a
-            # *non-interactive* SIGINT is forwarded correctly.
-            self.process.send_signal(SIGINT if not WINDOWS else SIGTERM)
-
     def returncode(self):
         if self.using_pty:
-            return os.WEXITSTATUS(self.status)
+            # No subprocess.returncode available; use WIFEXITED/WIFSIGNALED to
+            # determine whch of WEXITSTATUS / WTERMSIG to use.
+            # TODO: is it safe to just say "call all WEXITSTATUS/WTERMSIG and
+            # return whichever one of them is nondefault"? Probably not?
+            # NOTE: doing this in an arbitrary order should be safe since only
+            # one of the WIF* methods ought to ever return True.
+            code = None
+            if os.WIFEXITED(self.status):
+                code = os.WEXITSTATUS(self.status)
+            elif os.WIFSIGNALED(self.status):
+                code = os.WTERMSIG(self.status)
+                # Match subprocess.returncode by turning signals into negative
+                # 'exit code' integers.
+                code = -1 * code
+            return code
+            # TODO: do we care about WIFSTOPPED? Maybe someday?
         else:
             return self.process.returncode
 
+    def stop(self):
+        # No explicit close-out required (so far).
+        pass
+
 
 class Result(object):
     """
     A container for information about the result of a command execution.
 
-    See individual attribute/method documentation below for details.
+    All params are exposed as attributes of the same name and type.
+
+    :param str stdout:
+        The subprocess' standard output.
+
+    :param str stderr:
+        Same as ``stdout`` but containing standard error (unless the process
+        was invoked via a pty, in which case it will be empty; see
+        `.Runner.run`.)
+
+    :param str command:
+        The command which was executed.
+
+    :param str shell:
+        The shell binary used for execution.
+
+    :param dict env:
+        The shell environment used for execution. (Default is the empty dict,
+        ``{}``, not ``None`` as displayed in the signature.)
+
+    :param int exited:
+        An integer representing the subprocess' exit/return code.
+
+    :param bool pty:
+        A boolean describing whether the subprocess was invoked with a pty or
+        not; see `.Runner.run`.
+
+    :param tuple hide:
+        A tuple of stream names (none, one or both of ``('stdout', 'stderr')``)
+        which were hidden from the user when the generating command executed;
+        this is a normalized value derived from the ``hide`` parameter of
+        `.Runner.run`.
+
+        For example, ``run('command', hide='stdout')`` will yield a `Result`
+        where ``result.hide == ('stdout',)``; ``hide=True`` or ``hide='both'``
+        results in ``result.hide == ('stdout', 'stderr')``; and ``hide=False``
+        (the default) generates ``result.hide == ()`` (the empty tuple.)
 
     .. note::
         `Result` objects' truth evaluation is equivalent to their `.ok`
@@ -1021,27 +1081,37 @@ class Result(object):
                 do_something()
             else:
                 handle_problem()
+
+        However, remember `Zen of Python #2
+        <http://zen-of-python.info/explicit-is-better-than-implicit.html#2>`_.
     """
-    # TODO: inherit from namedtuple instead? heh
-    def __init__(self, command, shell, env, stdout, stderr, exited, pty):
-        #: The command which was executed.
-        self.command = command
-        #: The shell binary used for execution.
-        self.shell = shell
-        #: The shell environment used for execution.
-        self.env = env
-        #: An integer representing the subprocess' exit/return code.
-        self.exited = exited
-        #: An alias for `.exited`.
-        self.return_code = exited
-        #: The subprocess' standard output, as a multiline string.
+    # TODO: inherit from namedtuple instead? heh (or: use attrs from pypi)
+    def __init__(
+        self,
+        stdout="",
+        stderr="",
+        command="",
+        shell="",
+        env=None,
+        exited=0,
+        pty=False,
+        hide=tuple(),
+    ):
         self.stdout = stdout
-        #: Same as `.stdout` but containing standard error (unless the process
-        #: was invoked via a pty; see `.Runner.run`.)
         self.stderr = stderr
-        #: A boolean describing whether the subprocess was invoked with a pty
-        #: or not; see `.Runner.run`.
+        self.command = command
+        self.shell = shell
+        self.env = {} if env is None else env
+        self.exited = exited
         self.pty = pty
+        self.hide = hide
+
+    @property
+    def return_code(self):
+        """
+        An alias for ``.exited``.
+        """
+        return self.exited
 
     def __nonzero__(self):
         # NOTE: This is the method that (under Python 2) determines Boolean
@@ -1054,7 +1124,11 @@ class Result(object):
         return self.__nonzero__()
 
     def __str__(self):
-        ret = ["Command exited with status {0}.".format(self.exited)]
+        if self.exited is not None:
+            desc = "Command exited with status {0}.".format(self.exited)
+        else:
+            desc = "Command was not fully executed due to watcher error."
+        ret = [desc]
         for x in ('stdout', 'stderr'):
             val = getattr(self, x)
             ret.append(u"""=== {0} ===
@@ -1062,6 +1136,13 @@ class Result(object):
 """.format(x, val.rstrip()) if val else u"(no {0})".format(x))
         return u"\n".join(ret)
 
+    def __repr__(self):
+        # TODO: more? e.g. len of stdout/err? (how to represent cleanly in a
+        # 'x=y' format like this? e.g. '4b' is ambiguous as to what it
+        # represents
+        template = "<Result cmd={0!r} exited={1}>"
+        return template.format(self.command, self.exited)
+
     @property
     def ok(self):
         """
@@ -1088,11 +1169,11 @@ def normalize_hide(val):
     if val in (None, False):
         hide = ()
     elif val in ('both', True):
-        hide = ('out', 'err')
-    elif val == 'stdout':
-        hide = ('out',)
-    elif val == 'stderr':
-        hide = ('err',)
+        hide = ('stdout', 'stderr')
+    elif val == 'out':
+        hide = ('stdout',)
+    elif val == 'err':
+        hide = ('stderr',)
     else:
         hide = (val,)
     return hide
